#include "pe.h"
#include "glio.h"
#include <malloc.h>

#define min(A,B) (A>B?B:A)

extern int	wrboxsobj(OBJ *,int, int, int, int, int);
extern int	wrcirclesobj(OBJ *,int, int, int, int);
extern int	wrlinesobj(OBJ *,int, int, int, int, int, int, int, int);
extern int	appendobj(GRP *, OBJ *);
extern GRP	*allocgrp();
extern OBJ	*crboxsobj(int, int, int, int, int, int, char *);
extern OBJ	*crcirclesobj(int, int, int, int);
extern OBJ	*crlinesobj(int, int, int, int, int, int, int, int);
extern GRP	*cutgrp(GRP *);
extern int	insgrp(GRP *, GRP *);
extern ICELL	*cuticell(GRP *, ICELL *);
extern int	insicell(GRP *, ICELL *, ICELL *);
extern int 	groupbox(GRP *);
extern		selobj(int);

static	pushgrp(GRP *);

#define	box2box(bp1,bp2)	memcpy(bp1,bp2,sizeof(BOXPOS))

/********************************
*	create functions	*
*********************************/

static getobjstring(OBJ *obj)
{ return getstring(obj->body.framebox.s); }

global	create_box(int type,int x1,int y1)
{
	int	x2, y2;
	GRP	*grp;
	OBJ	*obj;
	int	line;

	gridmouse(&x1,&y1);
	if (drawing_box(x1,y1,&x2,&y2)==CRNONE)
		return CRNONE;

	obj = crboxsobj(type,x1,y1,x2,y2,5,"");
	grp=allocgrp();
	appendobj(grp,obj);
	redrawgrp(grp);
	getobjstring(obj);
	return CRSUCCESS;
}

global	create_circle(int type,int x1,int y1)
{
	int	r=0;
	GRP	*grp;
	OBJ	*obj;

	gridmouse(&x1,&y1);
	if (drawing_circle(type,x1,y1,&r)==CRNONE)
		return CRNONE;

	obj = crcirclesobj(type,x1,y1,r);
	grp=allocgrp();
	appendobj(grp,obj);
	redrawgrp(grp);
	return CRSUCCESS;
}

global	create_line(int type,int x1,int y1)
{
	int	x2, y2, dir_x, dir_y, len;
	GRP	*grp;
	OBJ	*obj;

	gridmouse(&x1,&y1);
	if (drawing_line(type,x1,y1,&x2,&y2,&dir_x,&dir_y,&len)==CRNONE)
		return CRNONE;

	obj = crlinesobj(type,x1,y1,x2,y2,dir_x,dir_y,len);
	grp=allocgrp();
	appendobj(grp,obj);
	redrawgrp(grp);
	return CRSUCCESS;
}

/********************************
*	resize functions	*
*********************************/

static	resize_box(OBJ *obj,int mx,int my)
{
	int	dx,dy,type,x1,y1,x2,y2;

	dy = obj->body.framebox.h; /* framebox, makebox, dashbox */
	dx = obj->body.framebox.w; /* framebox, makebox, dashbox */
	type = obj->type;
	x1 = obj->x;
	x2 = x1+dx;
	y2 = obj->y;
	y1 = y2-dy;

	if (abs(mx-x1)+abs(my-y1)>=(dx+dy)/2) {
		if (drawing_box(x1,y1,&x2,&y2)==CRNONE)
			return CRNONE;
	}
	else {
		if (moving_box(x1,y1,x2-x1,y2-y1,&dx,&dy)==CRNONE)
			return CRNONE;
		x1 += dx;	y1 += dy;
		x2 += dx;	y2 += dy;
	}
	wrboxsobj(obj,type,x1,y1,x2,y2);
	return CRSUCCESS;
}

static	resize_circle(OBJ *obj,int mx,int my)
{
	int	type,x1,y1,dx,dy;
	int	r;

	type = obj->type;
	x1 = obj->x;
	y1 = obj->y;
	r = obj->body.circle.di/2;

	if (l2norm(mx-x1,my-y1)>=r/2) {
		if (drawing_circle(type,x1,y1,&r)==CRNONE)
			return CRNONE;
	}
	else {
		if (moving_circle(x1,y1,r,&dx,&dy)==CRNONE)
			return CRNONE;
		x1 += dx;	y1 += dy;
	}
	wrcirclesobj(obj,type,x1,y1,r);
	return CRSUCCESS;
}

static int	dist(int x1,int y1,int x2,int y2,int mx,int my)
{
	long	xd,yd,xm,ym,l,d;

	xd=(long)x2-(long)x1;	yd=(long)y2-(long)y1;
	xm=(long)mx-(long)x1;	ym=(long)my-(long)y1;

	d = yd * xm - xd * ym;
	d *= d;
	l = xd * xd + yd * yd;
	if (l==0)
		d = xm * xm + ym * ym;
	else
		d /= l;
	return (int) d;
}

static	resize_line(OBJ *obj,int mx,int my)
{
	int	type,x1,y1,x2,y2,dir_x,dir_y,len,dx,dy;

	type = obj->type;
	x1 = obj->x;
	y1 = obj->y;
	dir_x = obj->body.line.dir_x;
	dir_y = obj->body.line.dir_y;
	len = obj->body.line.len;
	linexy2(x1,y1,dir_x,dir_y,len,&x2,&y2);
	dx = abs(x2-x1);
	dy = abs(y2-y1);

	if (dist(x1,y1,x2,y2,mx,my)>grid*grid/4)
		return CRNONE;

	if (abs(mx-x1)+abs(my-y1)>=(dx+dy)/2) {
		if (drawing_line(type,x1,y1,&x2,&y2,&dir_x,&dir_y,&len)
			==CRNONE)
			return CRNONE;
	}
	else {
		if (moving_line(x1,y1,x2-x1,y2-y1,&dx,&dy)==CRNONE)
			return CRNONE;
		x1 += dx;	y1 += dy;
		x2 += dx;	y2 += dy;
	}
	wrlinesobj(obj,type,x1,y1,x2,y2,dir_x,dir_y,len);
	return CRSUCCESS;
}

static int	distance(OBJ *obj,int x,int y)
{
	int	type,x1,y1,x2,y2,dx,dy,r;

	type = obj->type;
	x1 = obj->x;
	y1 = obj->y;

	if ((type>=LINETYPE)&&(type<=ARROWTYPE)) {
		linexy2(x1,y1,
			obj->body.line.dir_x,
			obj->body.line.dir_y,
			obj->body.line.len,
			&x2,&y2);
		return dist(x1,y1,x2,y2,x,y);
	}

	if ((type>=FRAMEBOXTYPE)&&(type<=MAKEBOXTYPE)) {
		y2 = y1;
		y1 -= obj->body.framebox.h;
		x2 = x1 + obj->body.framebox.w;
		dx = l2norm(x1-x,y1-y);
		dy = l2norm(x2-x,y2-y);
	}
	else {	/* if circle type */
		r = obj->body.circle.di/2;
		dx = l2norm(x1-x,y1-y);
		dy = dx-r;
		dy = abs(dy);
	}
	return min(dx,dy);
}

static int	nearest_obj(int x,int y,GRP **retgrp,ICELL **reticl)
{
	GRP	*grp;
	ICELL	*icl;
	OBJ	*obj;
	BOXPOS	box;
	int	d,m=16*grid;

	*retgrp=NULL;
	*reticl=NULL;
	box.x1 = x-grid;	box.x2 = x+grid;
	box.y1 = y-grid;	box.y2 = y+grid;
	for (grp = scr.group; grp!=NULL; grp=grp->succ) {
		if (!(overbox(&box,&(grp->box))))
			continue;
		for (icl = grp->first; icl!=NULL; icl=icl->succ) {
			obj = icl->obj;
			if ((overbox(&box,&(obj->box)))&&
			    ((d = distance(obj,x,y))<m)) {
				*retgrp = grp;
				*reticl = icl;
				m = d;
			}
		}
	}
	if (*reticl!=NULL)
		return CRSUCCESS;
	else
		return CRFAILURE;
}

typedef	int	(*RESFUNC)(OBJ *, int, int);

static	resize_grp(GRP *grp,ICELL *icl,int mx,int my)
{
	OBJ	*obj;
	ICELL	*succ;
	static RESFUNC	resizefunc[]={
		resize_box,
		resize_box,
		resize_box,
		resize_line,
		resize_line,
		resize_circle,
		resize_circle
	};

	obj = icl->obj;
	hideobjbx(obj);
	succ = cuticell(grp,icl);
	redrawbox(&(obj->box));
	(*resizefunc[obj->type-FRAMEBOXTYPE])(obj,mx,my);
	insicell(grp,icl,succ);
	redrawobj(obj);
	groupbox(grp);
}

global	resizegrp(int x,int y)
{
	GRP	*grp;
	ICELL	*icl;

	if (nearest_obj(x,y,&grp,&icl)==CRFAILURE)
		return;
	resize_grp(grp,icl,x,y);
}

/********************************
*	Move functions		*
*********************************/

static	moveobj(OBJ *obj,int dx,int dy)
{
	obj->x+=dx;
	obj->y+=dy;
	obj->box.x1+=dx;
	obj->box.x2+=dx;
	obj->box.y1+=dy;
	obj->box.y2+=dy;
}

global	movegrp_aux(GRP *grp,int dx,int dy)
{
	ICELL	*st;

	grp->box.x1+=dx;
	grp->box.y1+=dy;
	grp->box.x2+=dx;
	grp->box.y2+=dy;
	for (st = grp->first; st!=NULL; st=st->succ)
		moveobj(st->obj,dx,dy);
}

global	movegrp(int x,int y)
{
	GRP	*grp,*succ,*grp_over();
	BOXPOS	box;
	int	ret;

	box.x1 = box.x2 = x;
	box.y1 = box.y2 = y;
	grp = grp_over(&box);
	if (grp != NULL) {
		hidegrp(grp);
		succ = cutgrp(grp);

		if ((ret=moving_grpbox(x,y,&(grp->box),&x,&y))==CRSUCCESS)
			movegrp_aux(grp,x,y);

		insgrp(grp,succ);
		redrawgrp(grp);
		if (ret==CRSUCCESS)
			return CRSUCCESS;
	}
}

/********************************
*	Copy functions		*
*********************************/

static OBJ	*copyobj(OBJ *org)
{
	int	n;
	int	type = org->type;
	OBJ	*obj,*allocobj();

	if ((type>=FRAMEBOXTYPE)&&(type<=MAKEBOXTYPE))
		n = org->body.framebox.n;
	else
		n = 0;

	if ((obj = allocobj(n))==NULL)
		return NULL;
	memcpy(	(char *)obj,(char *)org,(int)sizeof(OBJ)+n);
	return obj;
}

static	copyicellchain(GRP *grp,ICELL *st)
{
	ICELL	*succ=NULL,*allocicell();

	while (st!=NULL) {
		succ = allocicell();
		succ->obj = copyobj(st->obj);
		insicell(grp,succ,NULL);
		st = st->succ;
	}
}

static GRP	*copygrp_aux(GRP *org)
{
	GRP	*grp,*allocgrp();

	grp=allocgrp();
	box2box(&(grp->box),&(org->box));
	copyicellchain(grp,org->first);
	return grp;
}

global	copygrp(int sx,int sy)
{
	GRP	*grp,*grp_over();
	BOXPOS	box;

	box.x1 = box.x2 = sx;
	box.y1 = box.y2 = sy;
	grp = grp_over(&box);
	if (grp != NULL) {
		if (moving_grpbox(sx,sy,&(grp->box),&sx,&sy)==CRSUCCESS) {
			if ((grp=copygrp_aux(grp))!=NULL) {
				movegrp_aux(grp,sx,sy);
				redrawgrp(grp);
			}
		}
	}
}

/********************************
*	Delete function		*
*********************************/

global	deletegrp(int x,int y)
{
	GRP	*grp,*grp_over();
	BOXPOS	box;
	XEvent ev;

	box.x1 = box.x2 = x;
	box.y1 = box.y2 = y;
	grp = grp_over(&box);
	if (grp != NULL) {
		hidegrp(grp);
		sel_grpbx(grp);

		for (;;) {
			do{
				XNextEvent(XtDisplay(Draw),&ev);
			}while(ev.type != ButtonPress);

			if (ev.xbutton.button == 3) {
				hidegrp(grp);
				redrawgrp(grp);
				return;
			}
			if (ev.xbutton.button == 1) {
				pushgrp(grp);
				return;
			}
		}
	}
}

static	pushgrp(GRP *grp)
{
	GRP	*pred,*succ;
	int	n;

	hidegrp(grp);
	succ = cutgrp(grp);
	n=0;
	for (pred=succ=scr.delgrp; pred!=NULL; pred=pred->pred) {
		succ = pred;
		n++;
	}
	if (n>=16) {
		pred=succ;
		succ=pred->succ;
		freeicellchain(pred->first);
		free((char *)pred);
		succ->pred=NULL;
	}
	grp->pred = scr.delgrp;
	grp->succ = NULL;
 	if(scr.delgrp != NULL) scr.delgrp->succ = grp;
	scr.delgrp = grp;
}

void undelgrp(Widget w,XtPointer client_data,XtPointer call_data)
{
	GRP	*pred,*succ,*grp;

	if ((grp=scr.delgrp)==NULL) return;
	selobj(UNDEL);
	scr.delgrp = scr.delgrp->pred;
	if (scr.delgrp != NULL) scr.delgrp->succ = NULL;

	insgrp(grp,NULL);
	redrawgrp(grp);
	selobj(mode);
}

/********************************
*	Edit functions		*
*********************************/

static	editobj(OBJ *obj,int x,int y)
{
	int	line,ret;

	if (inbox(x,y,obj->box)) {
		if (((ret=obj->type)>=FRAMEBOXTYPE)&&(ret<=MAKEBOXTYPE)) {
			hideobjbx(obj);
			sel_objbx(obj);
			ret = getobjstring(obj);
			hideobjbx(obj);
			redrawbox(&(obj->box));
			if (ret==0)
				return CRSUCCESS;
			else
				return CRFAILURE;
		}
	}
	return CRNONE;
}

static putlasticl(GRP *grp,ICELL *icl)
{
	cuticell(grp,icl);
	insicell(grp,icl,NULL);
}

static	editgrp_aux(GRP *grp,int x,int y)
{
	int	ret;
	ICELL	*icl;

	for (icl = grp->first; icl!=NULL; icl = icl->succ) {
		if ((ret=editobj(icl->obj,x,y))==CRNONE)
			continue;
		else if (ret == CRFAILURE)
			putlasticl(grp,icl);
		break;
	}
	return ret;
}

global	editgrp(int x,int y)
{
	GRP	*grp,*grp_over();
	BOXPOS	box;

	box.x1 = box.x2 = x;
	box.y1 = box.y2 = y;
	grp = grp_over(&box);
	if (grp != NULL)
		editgrp_aux(grp,x,y);
}

/************************
*	Grouping	*
*************************/

static GRP	*freeonlygrp(GRP *grp)
{
	GRP	*pred,*succ;

	succ = cutgrp(grp);
	free((char *)grp);
	return succ;
}

global	grouping(int x1,int y1)
{
	int	x2,y2;
	BOXPOS	box;
	GRP	*grp,*allocgrp();
	ICELL	*first=NULL, *last=NULL;

	if (selecting_box(x1,y1,&x2,&y2)==CRNONE)
		return CRNONE;

	boxpos(&box,x1,y1,x2,y2);
	for (grp=scr.group; grp!=NULL;) {
		if (!(boxinbox(grp->box,box))) {
			grp = grp->succ;
			continue;
		}
		if (first==NULL) {
			first = grp->first;	last = grp->last;
		}
		else {
			last->succ = grp->first;
			grp->first->pred = last;
			last = grp->last;
		}
		grp = freeonlygrp(grp);
	}
	if (first==NULL)
		return CRNONE;
	grp = allocgrp();
	grp->first = first;
	grp->last  = last;
	groupbox(grp);
}

global	ungroup(int x,int y)
{
	BOXPOS	box,*bp;
	GRP	*new,*grp,*allocgrp(),*grp_over();
	ICELL	*icell,*succ;

	box.x1 = box.x2 = x;
	box.y1 = box.y2 = y;
	grp = grp_over(&box);
	if (grp == NULL)
		return;
	for (icell=grp->first; icell!=NULL;) {
		new = allocgrp();
		bp = &(icell->obj->box);
		box2box(&(new->box),bp);
		new->first=new->last=icell;
		succ = icell->succ;
		icell->pred = icell->succ = NULL;
		icell = succ;
	}
	freeonlygrp(grp);
}
