#include "pe.h"
#include "glio.h"
#include <string.h>

extern GRP	*allocgrp();
extern OBJ	*crboxsobj(int, int, int, int, int, int, char *);
extern OBJ	*crcirclesobj(int, int, int, int);
extern OBJ	*crlinesobj(int, int, int, int, int,int, int, int);
extern int	appendobj(GRP *, OBJ *);
extern		selobj(int);
extern		redrawall();

static GRP	*newgrp;

static FILE	*fp;

static int	scr_w = 512;
static int	scr_h = 400;

#define	_others		-1
#define	_circle		0
#define	_dashbox	1
#define	_endpicture	2
#define	_framebox	3
#define	_line		4
#define	_makebox	5
#define	_picture	6
#define	_put		7
#define	_unitlength	8
#define	_vector		9
#define	keywordmax	10

static char	*keyword[] = {
	"circle",
	"dashbox",
	"endpicture",
	"framebox",
	"line",
	"makebox",
	"picture",
	"put",
	"unitlength",
	"vector"
};

static int	gridtrans(int x)
{
	x *= grid;	x /= oldgrid;
	return x;
}

static int	skip_line()
{
	unsigned char	c;

	while ((c=fgetc(fp))!=(unsigned char)EOF)
		if (c=='\n')
			break;
}

static unsigned char	skip_space()
{
	unsigned char	c;

	while ((c=fgetc(fp))!=(unsigned char)EOF) {
		if (c=='%')
			skip_line();
		else if ((c!=' ')&&(c!='\n')&&(c!='\t'))
			break;
	}
	return c;
}

static unsigned char	skip_until(unsigned char ch)
{
	unsigned char	c;

	while ((c=fgetc(fp))!=(unsigned char)EOF) {
		if (c==ch)
			break;
		if (c=='%')
			skip_line();
	}
	return c;
}

static unsigned char	get_int(int *xp)
{
	unsigned char	c;
	int	sig=1;
	long	n=0;

	c = skip_space();
	if (c=='-') {
		sig = -1;
		c = skip_space();
	}
	while ((c>='0')&&(c<='9')) {
		n*=10L;
		n+=(long)(c-'0');
		if (n>32767L) {
			n = 0;
			break;
		}
		c = fgetc(fp);
	}
	*xp = sig*(int)n;
	if ((c==' ')||(c=='\n')||(c=='\t'))
		c = skip_space();
	return c;
}

/* ({) n } */
static int	get_npar(int *np)
{
	if (get_int(np)=='}')
		return 0;
	else
		return -1;
}

/* ([) l|r b|t ] */
static int	get_pospar(int *np)
{
	unsigned char	c;
	int		lr=0, bt=0;

	for (;;) {
		c = skip_space();
		if	(c=='l')	lr = -1;
		else if (c=='r')	lr = 1;
		else if (c=='b')	bt = -1;
		else if (c=='t')	bt = 1;
		else if (c==']') {
			*np = 5 + bt * 3 + lr;
			return 0;
		}
		else
			return -1;
	}
}

/* {(} x , y ) */
static int	get_xypar(int *xp,int *yp)
{
	unsigned char	c;

	c = get_int(xp);
	if (c==',') {
		c = get_int(yp);
/* ( */		if (c==')')
			return 0;
	}
	return -1;
}

/* ({) text..{..}..} */
static int	get_textpar(char *s)
{
	unsigned char	c='\0';
	int	n,i;

	for (n=i=0; n>=0; s[i++]=c) {
		c = fgetc(fp);
		if (iskanji(c)) {
			s[i++]=c;
			c = fgetc(fp);
			continue;
		}
		if (c=='%') {
			skip_line();	continue;
		}
		if (c=='{')		n++;
		else if (c=='}')	n--;
	}
	s[i-1]='\0';
}

/* get keyword (only small letters) */

static int	getword(char *w)
{
	unsigned char	c;

	while (((c=fgetc(fp))>='a')&&(c<='z')) {
		*w++=c;
	}
	ungetc(c,fp);
	*w='\0';
}

static int	get_token()
{
	char	w[16];
	int	low,high,mid,cond;

	getword(w);
	low = 0;
	high = keywordmax-1;
	while (low <= high) {
		mid = (low+high)/2;
		if ((cond=strcmp(w,keyword[mid])) < 0)
			high = mid-1;
		else if (cond > 0)
			low = mid+1;
		else
			return mid;
	}
	return -1;
}

/********************************
*	Making Primitives	*
*********************************/

static int	makeboxs(int tok,int x,int y)
{
	OBJ		*obj;
	int		w,h,pos,n,flag;
	int		type, x1, y1, x2, y2;
	char		s[256];
	unsigned char	c;

	if (tok==_framebox)	type = FRAMEBOXTYPE;
	else if (tok==_makebox)	type = MAKEBOXTYPE;
	else if (tok==_dashbox)	type = DASHBOXTYPE;
	else
		return;

	flag = 0x00FF;
	pos = 5;
	c = skip_space();
	if ((type==DASHBOXTYPE)&&(c=='{')) {	/* } */
		get_npar(&n);
		c = skip_space();
	}
	if (c=='(') {	/* ) */
		if (get_xypar(&w,&h)==0)
			flag &= 0x000F;	/* w & h must be input */
		w = gridtrans(w);
		h = gridtrans(h);
		c = skip_space();
	}
	if (c=='[') {	/* ] */
		get_pospar(&pos);
		c = skip_space();
	}
	if (c=='{') {	/* } */
		get_textpar(s);
		flag &= 0x00F0;	/* text{} must be input */
	}

	if (flag == 0) {
		x1 = x;
		y1 = scr_h-y;
		x2 = x1+w;
		y2 = y1-h;
		obj = crboxsobj(type,x1,y1,x2,y2,pos,s);
		appendobj(newgrp,obj);
/**		group->first->obj->lattr = n;
*/	}
}

static int	makecircles(int tok,int x,int y)
{
	OBJ		*obj;
	int		type,x1,y1;
	int		di;
	unsigned char	c;

	if (tok!=_circle)
		return;

	c = skip_space();
	type = CIRCLETYPE;
	if (c=='*') {
		type = CIRCSTTYPE;
		c = skip_space();
	}
	if (c=='{') {
		get_npar(&di);
		if (type == CIRCLETYPE)
			di = gridtrans(di);
		x1 = x;
		y1 = scr_h-y;
		obj = crcirclesobj(type,x1,y1,di/2);
		appendobj(newgrp,obj);
	}
}

static int	makelines(int tok,int x,int y)
{
	OBJ	*obj;
	int	type, x1, y1, x2, y2;
	int	dir_x,dir_y;
	int	len;

	if	(tok==_line)	type = LINETYPE;
	else if (tok==_vector)	type = ARROWTYPE;
	else
		return;

	if ((skip_space()=='(')&&(get_xypar(&dir_x,&dir_y)==0)
	  &&(skip_space()=='{')) {
		get_npar(&len);
		len = gridtrans(len);
		x1 = x;		y1 = scr_h-y;
		linexy2(x1,y1,dir_x,dir_y,len,&x2,&y2);
		obj = crlinesobj(type,x1,y1,x2,y2,dir_x,dir_y,len);
		appendobj(newgrp,obj);
	}
}

/* read makebox, framebox, line, vector, or circle */
static int	read_prim()
{
	int	x,y;
	int	tok;

	if ((skip_space()=='(')&&(get_xypar(&x,&y)==0)
	  &&(skip_space()=='{')&&(skip_space()=='\\')) {
		x = gridtrans(x);
		y = gridtrans(y);
		tok = get_token();
		if ((tok==_makebox)||(tok==_framebox)||(tok==_dashbox))
			makeboxs(tok,x,y);
		else if ((tok==_line)||(tok==_vector))
			makelines(tok,x,y);
		else if (tok==_circle)
			makecircles(tok,x,y);
		if (skip_space()=='}')
			return 0;
	}
	return -1;
}

static int	read_pic()
{
	int	tok;

	for (;;) {
		if (skip_until('\\')==(unsigned char)EOF)
			return -1;
		if ((tok=get_token())==_picture) {
			if ((skip_space()=='(')	/* ) */
			  &&(get_xypar(&scr_w,&scr_h)==0))
				scr_w = gridtrans(scr_w);
				scr_h = gridtrans(scr_h);
				break;
		}
	}
	for (;;) {
		if (skip_until('\\')==(unsigned char)EOF)
			return -1;
		if ((tok=get_token())==_put)
			read_prim();
		else if (tok==_endpicture)
			return 0;
	}
}

global	finsub(char *s,int fl)
/* fl: if TRUE, fin without asking position */
{
	int	sx,sy;
	int	fivegrid = 5 * grid;
	Position	drx1,dry1;

	if (fl == FALSE) redrawall();
	if ((fp=fopen(s,"r"))==NULL)
		return;
	newgrp = allocgrp();
	read_pic();
	if (newgrp->first==NULL)
		freegrp(newgrp);
	else {
		{
			BOXPOS	*bp = &(newgrp->box);

			sx=fivegrid+(bp->x1+bp->x2)/2;
			sy=fivegrid+(bp->y1+bp->y2)/2;
		}
		if(fl == TRUE) movegrp_aux(newgrp,sx,sy);
		else{
			XtVaGetValues(Draw, XtNx, &drx1, XtNy, &dry1, NULL);
			if (sx>((int)drx1 + (int)drwidth))
				sx= (int)drx1 + ((int)drwidth)/2;
			if (sy>((int)dry1 + (int)drheight))
				sy= (int)dry1 + ((int)drheight)/2;
			place_mouse(sx,sy);
			movegrp_aux(newgrp,org_x+fivegrid,org_y+fivegrid);
			sx += org_x;	sy += org_y;
			if (moving_grpbox(sx,sy,&(newgrp->box),&sx,&sy)
					== CRSUCCESS) movegrp_aux(newgrp,sx,sy);
			redrawgrp(newgrp);
		}
	}
	fclose(fp);
}

void filein(Widget w,XtPointer client_data,XtPointer call_data)
{
	char	s[256];

	selobj(FILEI);
	s[0]='\0';
	getfilename(s);
	if (s[0]=='\0'){
		selobj(mode);
		return;}
	finsub(s,FALSE);
	strcpy(scr_filename,s);
	selobj(mode);
}
