#include "pe.h"
#include <malloc.h>

#define max(A,B) A>B?A:B
#define min(A,B) A>B?B:A

static	arrowbox(BOXPOS *,int,int,int,int);

global GRP	*cutgrp(GRP *grp)
{
	GRP	*pred = grp->pred;
	GRP	*succ = grp->succ;

	if (pred==NULL)		scr.group = succ;
	else			pred->succ = succ;
	if (succ!=NULL)		succ->pred = pred;
	return succ;
}

global insgrp(GRP *grp,GRP *succ)
{
	GRP	*pred;

	if (succ==NULL)
		for (pred=succ=scr.group; succ!=NULL; succ=succ->succ)
			pred = succ;
	else {
		pred = succ->pred;
		succ->pred = grp;
	}
	if (pred==NULL)		scr.group = grp;
	else			pred->succ = grp;
	grp->pred = pred;
	grp->succ = succ;
}

global ICELL	*cuticell(GRP *grp,ICELL *icl)
{
	ICELL	*pred = icl->pred;
	ICELL	*succ = icl->succ;

	if (pred==NULL)		grp->first = succ;
	else			pred->succ = succ;
	if (succ==NULL)		grp->last = pred;
	else			succ->pred = pred;
	return succ;
}

global insicell(GRP *grp,ICELL *icl,ICELL *succ)
{
	ICELL	*pred;

	if (succ==NULL) {
		pred = grp->last;
		grp->last = icl;
	}
	else {
		pred = succ->pred;
		succ->pred = icl;
	}
	if (pred==NULL)		grp->first = icl;
	else			pred->succ = icl;
	icl->succ = succ;
	icl->pred = pred;
}

global 	groupbox(GRP *grp)
{
	int	x1,y1,x2,y2;
	ICELL	*icl;
	BOXPOS	*bp;

	if (grp==NULL)			return CRNONE;
	if ((icl=grp->first)==NULL)	return CRNONE;

	bp = &(icl->obj->box);
	x1 = bp->x1;
	y1 = bp->y1;
	x2 = bp->x2;
	y2 = bp->y2;

	for (icl=icl->succ; icl!=NULL; icl=icl->succ) {
		int	z;
		bp = &(icl->obj->box);
		if ((z = bp->x1)<x1)	x1 = z;
		if ((z = bp->y1)<y1)	y1 = z;
		if ((z = bp->x2)>x2)	x2 = z;
		if ((z = bp->y2)>y2)	y2 = z;
	}
	boxpos(&(grp->box),x1,y1,x2,y2);
}

global	GRP *allocgrp()
{
	GRP	*grp,*last,*pred;

	grp = (GRP *)malloc((int)sizeof(GRP));
	insgrp(grp,NULL);
	grp->first=grp->last=NULL;
	return (grp);
}

global	ICELL *allocicell()
{
	ICELL	*icell;

	icell = (ICELL *)malloc(sizeof(ICELL));
	icell->pred = icell->succ = NULL;
	return (icell);
}

global	OBJ *allocobj(int n)
{
	return (OBJ *)malloc((int)sizeof(OBJ)+n);
}

global	freeicellchain(ICELL *st)
{
	ICELL	*succ;

	while (st!=NULL) {
		succ = st->succ;
		free((char *)st->obj);
		free((char *)st);
		st = succ;
	}
}

global	freegrp(GRP *grp)
{
	freeicellchain(grp->first);
	cutgrp(grp);
	free((char *)grp);
}

global	int	appendobj(GRP *grp,OBJ *obj)
{
	ICELL	*icell;

	icell=allocicell();
	icell->obj=obj;
	insicell(grp,icell,NULL);
	groupbox(grp);
}

global	int	wrboxsobj(OBJ *obj,int type,int x1,int y1,int x2,int y2)
{
	int	minx,maxx,midx;

	obj->type = type;

	if (x1<x2)	minx = x1,maxx=x2;	else minx = x2, maxx = x1;
	midx = (x1+x2)/2-4;
	obj->box.x1 = min(minx,midx);
	midx += 8;
	obj->box.x2 = max(maxx,midx);
	obj->x = minx;

	if (y1<y2)	minx = y1,maxx=y2;	else minx = y2, maxx = y1;
	midx = (y1+y2)/2-4;
	obj->box.y1 = min(minx,midx);
	midx += 8;
	obj->box.y2 = max(maxx,midx);
	obj->y = maxx;

	x1 -= x2;	obj->body.framebox.w=abs(x1);
	y1 -= y2;	obj->body.framebox.h=abs(y1);
}

global	OBJ	*crboxsobj(int type,int x1,int y1,int x2,int y2,int pos,char *s)
{
	OBJ	*obj,*allocobj();

	if ((obj=allocobj(MAXLEN))==NULL)
		return NULL;
	wrboxsobj(obj,type,x1,y1,x2,y2);
	obj->body.framebox.pos=pos;
	obj->body.framebox.n=MAXLEN;
	memcpy(obj->body.framebox.s,s,MAXLEN);
	return obj;
}

global	int	wrcirclesobj(OBJ *obj,int type,int x1,int y1,int r)
{
	{
		BOXPOS	*bp = &(obj->box);

		bp->x1 = x1-r;
		bp->y1 = y1-r;
		bp->x2 = x1+r;
		bp->y2 = y1+r;
	}
	obj->type = type;
	obj->x = x1;
	obj->y = y1;
	obj->body.circle.di = 2*r;
}

global	OBJ	*crcirclesobj(int type,int x1,int y1,int r)
{
	OBJ	*obj,*allocobj();

	if ((obj=allocobj(0))==NULL)
		return NULL;
	wrcirclesobj(obj,type,x1,y1,r);
	return obj;
}

global	int	wrlinesobj(OBJ *obj,int type,int x1,int y1,int x2,int y2,
			int dir_x,int dir_y,int len)
{
	obj->type = type;
	obj->x = x1;
	obj->y = y1;
	obj->body.line.dir_x=dir_x;
	obj->body.line.dir_y=dir_y;
	obj->body.line.len = len;
	if (type==ARROWTYPE)
		arrowbox(&(obj->box),x1,y1,x2,y2);
	else
		boxpos(&(obj->box),x1,y1,x2,y2);
}

global	OBJ	*crlinesobj(int type,int x1,int y1,int x2,int y2,
				int dir_x,int dir_y,int len)
{
	OBJ	*obj,*allocobj();

	if ((obj=allocobj(0))==NULL)
		return NULL;
	wrlinesobj(obj,type,x1,y1,x2,y2,dir_x,dir_y,len);
	return obj;
}

/********************************
*	line functions		*
*********************************/

static arrowrlsub(int tate,int yoko,int nagasa)
{
	if (nagasa<=0)
		return 0;
	else
		return -(tate * 8 + yoko * 4)/nagasa;
}

global	arrowrl(int x1,int y1,int x2,int y2,int *xr,int *yr,int *xl,int *yl)
{
	int	dir_x = x2 - x1;
	int	dir_y = y2 - y1;
	int	d;

	d = l2norm((int)dir_x,(int)dir_y);
	*xr = x2 + arrowrlsub(dir_x,-dir_y,d);
	*yr = y2 + arrowrlsub(dir_y, dir_x,d);
	*xl = x2 + arrowrlsub(dir_x, dir_y,d);
	*yl = y2 + arrowrlsub(dir_y,-dir_x,d);
}

static	arrowbox(BOXPOS *mmbox,int x1,int y1,int x2,int y2)
{
	int	xr,yr,xl,yl;
	int	minx,miny,maxx,maxy;

	arrowrl(x1,y1,x2,y2,&xr,&yr,&xl,&yl);
	if (x1<x2)	minx = x1, maxx = x2;	else minx = x2, maxx = x1;
	if (xr<xl)	miny = xr, maxy = xl;	else miny = xl, maxy = xr;
	mmbox->x1 = min(minx,miny);
	mmbox->x2 = max(maxx,maxy);
	if (y1<y2)	minx = y1, maxx = y2;	else minx = y2, maxx = y1;
	if (yr<yl)	miny = yr, maxy = yl;	else miny = yl, maxy = yr;
	mmbox->y1 = min(miny,minx);
	mmbox->y2 = max(maxy,maxx);
}

global gridint(int x)
{
	return ((x+grid/2)/grid)*grid;
}

global	linexy(int n,int x1,int y1,int mx,int my,
		int *dir_x,int *dir_y,int *x2,int *y2)
{
	int	i,j,d,minimum;
	int	_dir_x,_dir_y,_x2,_y2;

	_x2 = abs(mx-x1);
	_y2 = abs(my-y1);
	if (_x2 > _y2) {
		minimum = _y2/2;
		_dir_x = 1; _dir_y = 0;
	} else {
		minimum = _y2/2;
		_dir_x = 0; _dir_y = 1;
	}
	for (i=1; i<n; i++) {
		for (j=1; j<n; j++) {
			d = abs(_y2 - j * _x2 / i);
			if (d<minimum) {
				_dir_x = i;
				_dir_y = j;
				minimum = d;
			}
		}
	}
	_dir_x *= sign((mx-x1));
	_dir_y *= sign((y1-my));
	if (_dir_x==0) {
		_x2 = x1;
		_y2 = my;
	}
	else {
		_x2 = mx;
		_y2 = y1-_dir_y*(mx-x1)/(_dir_x);
	}

	if (_dir_x==0)
		_y2 = gridint(_y2);
	if (_dir_y==0)
		_x2 = gridint(_x2);
	*dir_x	= _dir_x;
	*dir_y	= _dir_y;
	*x2	= _x2;
	*y2	= _y2;
}

global	linexy2(int x1,int y1,int dir_x,int dir_y,int len,int *x2,int *y2)
{
	if (dir_x==0) {
		*x2 = x1;
		*y2 = y1 - sign(dir_y) * len;
	}
	else {
		*x2 = x1 + sign(dir_x) * len;
		*y2 = y1 - sign(dir_x) * len * dir_y / dir_x;
	}
}

global	int gridmouse(int *x,int *y)
{
	*x = gridint(*x);
	*y = gridint(*y);
}

global int	l2norm(int x, int y)
{
	unsigned int	high,low,mid;
	unsigned long	xx,tmp;

	x = abs(x);
	y = abs(y);
	xx = (long)x * (long)x + (long)y * (long)y;
	low = max(x,y);
	high = min(x,y);
	high += low;
	while (low<=high) {
		mid = (low+high)/2;
		tmp = (unsigned long) mid*mid;
		if (xx < tmp)
			high = mid-1;
		else if (xx > tmp)
			low = mid+1;
		else
			return (int)mid;
	}
	if (low>=mid)
		return (int)high;
	else
		return (int)(low+mid+1)/2;
}

global	boxpos(BOXPOS *bp,int x1,int y1,int x2,int y2)
{
	if (x1<x2)	bp->x1 = x1, bp->x2 = x2;
	else		bp->x1 = x2, bp->x2 = x1;
	if (y1<y2)	bp->y1 = y1, bp->y2 = y2;
	else		bp->y1 = y2, bp->y2 = y1;
}
